/**************************************************************************
 
 avih-commented.c
 
 ANOTHER VISIT IN HELL (commented version)
 by Jakub Wasilewski
 version 1.2
 
 This file is an attempt at writing a 1kB roguelike. After removing the
 comments and unnecessary whitespace, this source file is exactly 1024 bytes
 long.
 
 Parts of the source code are marked '-unsafe-'. That means that this part was
 done in a specific way in order to shorten the source code, and it might not
 work correctly on your system (though it should work on most 32-bit Linuxes). 
 If you want a version that should compile and work correctly everywhere 
 (including on Windows), use avih-safe.c.
 
 See README for playing instructions.

 This file is under the zlib/libpng license.

***************************************************************************/

/**************************************************************************
  Preprocessor definitions.
***************************************************************************/

/*-unsafe-: #includes for stdlib and curses are omitted. The functions
  used in the code work with implicit declarations on every system I tried 
  the game on - which doesn't make it correct, of course. */

#include <curses.h>

/* Shorter expression for a 'for' loop. */
/* One additional trick that we can use with that macro is creating a compound 
   termination condition by appending something to h, like 
   F(i,8&&(something else)). */
#define F(v,h)for(v=0;v<h;v++)
/* Process instructions 'i' for every tile in the dungeon. 'i' can contain
   semicolons (the last instruction is not required to have one). */
#define D(i)F(y,23)F(x,T){i;}
/* Short form for 'dungeon tile at position x,y'. */
#define G d[x][y]
/* Sign of number 'n' */
#define Z(n)(n>0)-(n<0)
/* Short form for dungeon tile at position x+sgn(f),y+sgn(g) - 
   used primarily by movement calculations. */
#define M d[x+Z(f)][y+Z(g)]
/* Item handling helper. Basically checks if the tile the player walked into
   is 'v', and if so, executes 'i', removing the tile afterwards. */
#define C(v,i)M^v||(i,M=e),
/* Short form for a random number, used as r%N for a random number between
   0 and N-1. */
#define r rand()

/**************************************************************************

	Variables:
	x,y - always indicate a position in the dungeon, used mostly through the 
	      G macro.
	f,g - in main loop - stores player position
	      in m() - stores movement direction (indirectly, [sgn(f),sgn(g)] is
	               the actual direction

	L   - current dungeon level
	l,p - HP current/maximum
	v   - gold collected
	j   - mana points
	a   - concentration
	w   - how many sides the weapon die has
	
	c   - stores the character pressed by the user (minus 1)
	b   - used as a loop counter and temporary storage
	      in LOS code - the colour of the tile
	      
	d[][] - dungeon map
	
	Constants:
	T   - the player tile, '@', also width of dungeon
	e   - empty tile, '.'
	u   - a constant used for monster handling, 'a'-1.
	      basically tile>u means that tile is a monster
	K   - 0x100, a bit mask - if this bit is set on a tile, the occupant of
	      this tile (player or monster) has already moved this turn
	      
	k[] - a multi-purpose table containing tiles, used for:
	      - coloring output: entries 0-7
	      - putting initial tiles in the dungeon: entries 2-11
          - generating loot: entries 9-13
          
***************************************************************************/
int x,y,f,g,	
	L,l=18,p=18,v,j,a,w=1,
	c,b,
	d[90][90],
	T=64,e=46,u=96,K=256
/*-unsafe-: initializing an int array with a wide string (L"...") is not
  guaranteed to work on systems where wchar_t != int, which is Windows and
  most 64-bit systems. */
;
wchar_t k[]=L"--'*$.@'>$'/'$";

/**************************************************************************
  Move the actor at (x,y) by [sgn(f),sgn(g)] 
***************************************************************************/
m(f,g){
	/* Is it the player moving? */
	G-T
	?
		/* Not a player, so it's a monster moving. */
		/* Monster moved into player? */
		M-T||
			/* Yes, let's hit him for some damage */
			(l-=G-u)>0||
				/* And kill him if his hit-points are down to 0. */
				(M=e)
	:
	(
		/* This IS the player moving. */
		/* Let's handle all the things he might bump into. */
		
		/* Souls */
		C(39,p+=L)
		/* Power gems */
		C(42,l=p)
		/* Gold */
		C(36,v+=L*L)
		/* Weapons */
		C(47,w=abs(L+r%4-3)+1)
		
		/* Monsters */
		M>u&&
			/* Hit it with the player's weapon. */
			(M-=r%w+a+1)%K<=u&&(
				/* Turns out it's dead: leave loot, give mana, and reset
				   movement. */
				M=k[r%5+9],
				j++,
				f=g=0
			),
			
		/* Stairs */
		M^62||N(),
		
		/* Move complete: the only thing left is determining if the PC has
		   concentration - only if he rested this turn. */
		a=c==52
	);
	
	/* All items and special features are handled now. It's time for actual
	   movement to happen. */
	/* Are we walking into an empty tile? */
	M-e
	?
		/* Nope, not empty. */
		/* Were we moving diagonally into something inanimate? */
		M<T&&f*g&&
			/* Yes - let's try moving horizontally or vertically instead. This
			   helps with making the AI less stupid - and we get Doom style 
			   sliding along walls for the player as a bonus. */
			(m(f,0),m(0,g))
	:
		/* Yes, we are walking into an empty tile. */
		/* Proceed with the actual move, mark the actor as having already
		   moved this turn. */
		(M=G|K,G=e);
}

/**************************************************************************
  Finds a random empty position in the dungeon and inserts 't' in that
  position.
***************************************************************************/
E(t){
	/* Generate random positions until an empty one is found. */
	while(x=r%T,y=r%23,G-e);
	/* Place 't' in the empty position. */
	G=t;
}

/**************************************************************************
  Generates a new dungeon level.
***************************************************************************/
N(){
	/* Initialize it with walls */
	D(G=35)
	/* Initialize position for the first corridor. */
	x=y=9;
	/* Increase level number. */	
	++L;
	/* Generate 21 corridors, each starting where the previous one ended. */
	F(b,21)
	{
		/* Generate an end position for the corridor. */
		f=r%61+2;
		g=r%20+1;
		/* This loop should really be a 'while' that terminates when the end
		   position is reached. The probability of not reaching it in 256
		   iterations is vanishingly small, so this for loop is a short-cut. */
		F(c,K)
			/* Drill an empty position at x,y. */
			G=e,
			/* Move x,y one tile towards the end position, choosing randomly 
			   between moving horizontally and moving vertically. */
			r%2?
				(x+=Z(f-x))
			:
				(y+=Z(g-y));
	}
	/* This is required so that the fake level 9 that does not contain the
	   player is not drawn on-screen. */
	f=0;
	
	/* Put pre-defined objects from the k[] table into the dungeon. */
	F(b,9)
		/* But not on level 9, which is fake. */
		L<9&&
			/* g%2 is a random offset into the table that makes half the
			   levels contain a weapon, while the other half contains an
			   additional soul to grab. */
			E(k[b+g%2+2]);
	
	/* Put monsters in the dungeon, with some variation in terms of number. */
	F(b,L*2+g%3)
		/* The power of each monster is randomly determined based on the
		   level. 9/5 is a balancing constant and can be altered to get
		   a harder or easier game. */
		E(99+r%(L*9/5));
	
	/* Clear the screen. The game uses the screen itself as "dungeon memory",
	   so this effectively clears it. */
	clear();
}

/**************************************************
  Main function
***************************************************/

main()
{
	/** Initialization **/
	/* Curses initialization coupled with seeding the RNG. */
	/*-unsafe-: on some systems, initscr() can return the same value every
	    time the game is launched, which makes for a very poor seed indeed. */
	srand(initscr());
	raw();
	start_color();
	
	/* Set up curses color pairs. */
	int q[]={COLOR_BLACK,COLOR_RED,COLOR_GREEN,COLOR_YELLOW,COLOR_BLUE,COLOR_MAGENTA,COLOR_CYAN,COLOR_WHITE};
	F(b,8)
		/*-unsafe-: the numeric values for colours are different under Linux
		  ncurses and Windows pdcurses. This will cause the game to be coloured
		  very differently (and garishly) under Windows. */
		init_pair(b+1,q[b],0);

	/* Generate the first level. */
	N();
	
	/** Main loop **/
	/* Play until 'Q' is pressed. */
	while(c-80)
	{
		/* Clear move flags, find player, store his position in f,g. */
		D(
			G%=K;
			G-T||
				(f=x,g=y)
		)
		
		/* Draw screen. */
		D(
			/* Determine the colour the tile will have, if visible. The k[]
			   table is used as an index of tile colours. If something is not
			   found, then it is either a monster (and will have colour 7), or
			   a gray tile (colour 8, stairs, weapons, empty tiles). */
			F(b,8-G/u&&G^k[b]);
			/* Check for tile visibility. This loop examines tiles on the
			   ray that connects f,g (the player's position) to x,y. If a
			   wall is found, the colour for the tile is set to 1 (which
			   happens to be black on black). */
			F(c,u)
				d[f+(x-f)*c/u][g+(y-g)*c/u]-35||
					(b=1);
					
			/* Draw a tile if any of the following conditions are met:
			   - empty tiles are drawn whether visible or not - this makes sure
			     empty tiles are not stored in the "dungeon memory".
			   - monsters are drawn whether visible or not - again, to make
			     sure they don't leave 'shadows' in the "dungeon memory".
			   - tiles the PC sees are drawn - those are refreshed with the new
			     contents
			   - tiles in the first column are always drawn - this ensures that
			     the status bar ends up in the proper row, since the last tile
			     printed is always at (0,22).
			*/
			G!=e&G<u&b<2&&x||
				/*-unsafe-: b*K is what COLOR_PAIR(b) expands to on every
				  system I had access to, which doesn't mean it's correct
				  everywhere. */
				mvaddch(y,x,G|COLOR_PAIR(b))
		)
		
		/* Draw status bar. */
		printw("\nHP:%d/%d M:%d W:d%d L%d $%d %c ",l,p,j,w,L,v,L^9?32+a:87);
		
		/* Get input from player. */
		/* One is subtracted from the input to make calculations shorter. */
		c=mvgetch(g,f)-1;
		
		/* Move all monsters. */
		D(
			/* Is this tile a monster? */
			G/u^1||
				/* Yes it is - move it towards the player's position. */
				m(f-x,g-y)
		)

		/* Move the player. */
		D(
			/* Is this tile a player? */
			G-T||(
				/* Is this a special move (teleport, 'a' was pressed)? */
				c^u?
					/* No, it's a normal move. */
					/* Extract direction from 'c' and move the player. */
					m(c%3-1,17-c/3)
				:
					/* Yes, it's a teleport. */
					/* Do we have enough mana for this? */
					j>5&&(
						/* Yes, we do. Move the player to a new, random
						   position and deduct the mana cost. */
						j-=6,
						G=e,
						E(T)
					)
			)
		)
	}
	
	/** Deinitialization **/
	/* Curses clean up to restore default terminal settings. */
	endwin();
}

